"""
A collection of serializers that can operate on dict types. plumbing
implementations can optionally use these serializers.
"""

import json


class JSON:
    """Provide serialize() and deserialize() methods to deal with JSON."""

    def serialize(self, pkg):
        """Return a string representation of the data."""
        return json.dumps(pkg)

    def deserialize(self, msg):
        """Return a dict from the JSON string."""
        return json.loads(msg)



"""
Top level utilities and documentation for the plumbing implementation in AWS.
"""

import boto3


class SQSResource:
    """Mixin to provide _get_queue() to fetch a SQS queue by name."""

    def _get_queue(self, queue_name, sqs=None):
        """Get the queue object by name.

        The SQS resource handle can be passed, otherwise one is instantiated.

        Positional paramters:
        * the queue name.

        Keyword parameters:
        * sqs: optionally the SQS resource handle.
        """
        if sqs is None:
            sqs = boto3.resource('sqs')
        return sqs.get_queue_by_name(QueueName=queue_name)


class SNSResource:
    """Mixin to provide _get_topic() to fetch a SNS topic by name."""

    def _get_topic(self, topic_name, sns=None):
        """Get the topic by name.

        A SNS resource handle can be passed, or one will be obtained with
        default configuration parameters.

        Positional parameters:
        * the topic name.
        * optionaly, a SNS handler to get the topic by name.
        """
        if sns is None:
            sns = boto3.resource('sns')
        # Use property of SNS resource: create_topic() is idempotent.
        # http://boto3.readthedocs.org/en/latest/reference/services/sns.html#sns.Client.create_topic
        return sns.create_topic(Name=topic_name)




# TODO: fix this doc to show the protocol, and the responsibilities of an
# implemnetation.
"""Fetch raw packages and push processed packages.

To fetch raw packages use a "Package Source" object, which handles a backend
connection and its configuration and exposes a "get()" method. Similarly, a
"Package Sink" object exposes a "put(pkg)" method.
"""

# TODO: create BaseSource and BaseSink.



"""
In-memory source and sink implementation.

Designed as a test artifact for external systems.
"""


class Source:
    """Test artifact, initialized with packages.

    It provides "replay" functionality by returning pre-loaded packages.
    """

    def __init__(self, packages):
        """Get a reference to the package repository.

        Positional parameter:
        * packages list.
        """
        self.depth = len(packages)
        self.packages = packages
        # Will use list.pop() to get the packages.
        self.packages.reverse()

    def get(self):
        """Pop first package from the repository.

        Raises RuntimeError if there are no more packages."""
        try:
            return self.packages.pop()
        except IndexError:
            raise RuntimeError('no more packages in MemorySource')


class Sink:
    """Wrap a list to simulate a Package Sink object."""

    def __init__(self):
        self.packages = []

    def put(self, pkg):
        self.packages.append(pkg)



from setuptools import setup

setup(
    name = 'plumb',
    packages = [
        'plumbing',
        'plumbing.aws',
        'plumbing.redis',
    ],
    version = '0.5.0',
    description = 'Connect systems via Redis, AWS SQS and SNS',

    install_requires=['boto3', 'redis'],
    tests_require=['awstestutils', 'mockredispy'],

    test_suite = 'tests.unit',

    author = 'Elvio Toccalino',
    author_email = 'elvio@spect.ro',

    keywords = ['redis', 'AWS', 'queues', 'distributed'],
    classifiers = [
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3',
        'Topic :: Utilities',
        'Topic :: System :: Networking',
        'Topic :: System :: Distributed Computing',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Object Brokering',
    ]
)



import logging
from . import SQSResource
from ..serializers import JSON


class Source(SQSResource, JSON):
    """Uses the boto3 Queue object.

    Uses get_queue_by_name(), or receives the queue during initialization step.

    Relies on receive_messages() as described in:
    http://boto3.readthedocs.org/en/latest/reference/services/sqs.html#queue
    """

    # http://boto3.readthedocs.org/en/latest/reference/services/sqs.html
    # http://docs.aws.amazon.com/AWSSimpleQueueService/latest/SQSDeveloperGuide/sqs-long-polling.html
    # https://github.com/boto/boto3-sample/blob/master/transcoder.py

    #
    # TODO: use queue name for logging.
    #

    # For reference value, see:
    # http://docs.aws.amazon.com/AWSSimpleQueueService/latest/SQSDeveloperGuide/sqs-long-polling.html
    LONG_POLLING_TIMEOUT = 20  # seconds

    def __init__(self, queue=None, queue_name=None, sqs=None):
        """Prepare the source with a configured SQS backend.

        The SQS queue object can be passed, or the queue name. If both are
        passed, the object is preferred.

        Keyword parameters:
        * queue: SQS client object.
        * queue_name: alternatively, the queue name.
        * sqs: optionally, a SQS resource handle can be passed.

        """
        self.log = logging.getLogger('sqs.source')
        if queue is None:
            queue = self._get_queue(queue_name, sqs=sqs)
        self.queue = queue
        self.log.debug('using SQS queue URL="%s"' % self.queue.url)

    def _package_from_message(self, msg):
        """Parse the raw package from the message coming from the backend."""
        try:
            return self.deserialize(msg.body)
        except Exception as e:
            self.log.error('could not parse incoming message')
            self.log.debug('offending message: %s' % msg.body)
            raise e

    def _get_message(self):
        """Returns a single (deleted) message from backend, or None.

        Uses queue.receive_messages() to poll for a package. Polls the backend
        until it gets a new message. Blocks for LONG_POLLING_TIMEOUT.

        Counts on LONG_POLLING_TIMEOUT seconds being a good polling timeout,
        and 1 being the default number of messages to pull.
        """
        # receive_messages() returns a list, even if asked for just one.
        msgs = self.queue.receive_messages(
            WaitTimeSeconds=self.LONG_POLLING_TIMEOUT,
        )
        try:
            msg = msgs[0]
        except IndexError:
            # No messages available.
            return None

        msg.delete()
        return msg

    def get_once(self):
        """Returns a package if one is available, or None."""
        msg = self._get_message()
        if msg is not None:
            try:
                msg = self._package_from_message(msg)
            except Exception:
                self.log.warning('dropping offending message')
                msg = None
        return msg

    def get(self, num_tries=-1):
        """Returns a package, or None, after num_tries times.

        Use "num_tries" to specify the number of loops. A value of -1 is an
        infinite loop.
        """
        count = num_tries
        while count != 0:
            pkg = self.get_once()
            if pkg is not None:
                return pkg
            if count > 0:
                count = count - 1
        return None


class Sink(SQSResource, JSON):
    """Uses the boto3 Queue object.

    Uses get_queue_by_name(), or receives the queue during initialization step.

    Relies on send_message() as described in:
    http://boto3.readthedocs.org/en/latest/reference/services/sqs.html#queue
    """

    # http://boto3.readthedocs.org/en/latest/reference/services/sqs.html
    # http://docs.aws.amazon.com/AWSSimpleQueueService/latest/SQSDeveloperGuide/sqs-long-polling.html
    # https://github.com/boto/boto3-sample/blob/master/transcoder.py

    def __init__(self, queue=None, queue_name=None, sqs=None):
        """Prepare the sink with a configured SQS backend.

        Either the SQS queue object or its name must be passed (queue object is
        preferred).

        Keyword parameters:
        * queue: the queue object.
        * queue_name: alternatively, the queue name.
        * sqs: optionaly, a SQS handler to get the queue by name.
        """
        self.log = logging.getLogger('sqs.sink')
        if queue is None:
            queue = self._get_queue(queue_name, sqs)
        self.queue = queue
        self.log.debug('using SQS queue URL="%s"' % self.queue.url)

    def put(self, pkg):
        """Sends the raw package to the queue.

        Positional parameters:
        * the raw package dict.
        """
        msg = self.serialize(pkg)
        self.log.debug('sending to queue JSON encoded package: %s' % msg)
        self.queue.send_message(MessageBody=msg)
